from . import VEXObject


class IRStmt(VEXObject):
    """
    IR statements in VEX represents operations with side-effects.
    """

    __slots__ = ['arch', 'tag']

    def __init__(self, c_stmt, irsb):
        VEXObject.__init__(self)
        self.arch = irsb.arch
        # self.c_stmt = c_stmt
        self.tag = ints_to_enums[c_stmt.tag]

    def pp(self):
        print self.__str__()

    @property
    def expressions(self):
        expressions = []
        for k in self.__slots__:
            v = getattr(self, k)
            if isinstance(v, IRExpr):
                expressions.append(v)
                expressions.extend(v.child_expressions)
        return expressions

    @property
    def constants(self):
        return sum((e.constants for e in self.expressions), [])

    @staticmethod
    def _translate(c_stmt, irsb):
        if c_stmt[0] == ffi.NULL:
            return None

        tag = c_stmt.tag
        try:
            stmt_class = _tag_to_class[tag]
        except KeyError:
            raise PyVEXError('Unknown/unsupported IRStmtTag %s\n' % ints_to_enums[tag])
        return stmt_class(c_stmt, irsb)


class NoOp(IRStmt):
    """
    A no-operation statement. It is usually the result of an IR optimization.
    """
    def __init__(self, c_stmt, irsb):  # pylint:disable=unused-argument
        IRStmt.__init__(self, c_stmt, irsb)

    def __str__(self):
        return "IR-NoOp"


class IMark(IRStmt):
    """
    An instruction mark. It marks the start of the statements that represent a single machine instruction (the end of
    those statements is marked by the next IMark or the end of the IRSB).  Contains the address and length of the
    instruction.
    """

    __slots__ = ['addr', 'len', 'delta']

    def __init__(self, c_stmt, irsb):
        IRStmt.__init__(self, c_stmt, irsb)
        self.addr = c_stmt.Ist.IMark.addr
        self.len = c_stmt.Ist.IMark.len
        self.delta = c_stmt.Ist.IMark.delta

    def __str__(self):
        return "------ IMark(0x%x, %d, %d) ------" % (self.addr, self.len, self.delta)


class AbiHint(IRStmt):
    """
    An ABI hint, provides specific information about this platform's ABI.
    """

    __slots__ = ['base', 'len', 'nia']

    def __init__(self, c_stmt, irsb):
        IRStmt.__init__(self, c_stmt, irsb)
        self.base = IRExpr._translate(c_stmt.Ist.AbiHint.base, irsb)
        self.len = c_stmt.Ist.AbiHint.len
        self.nia = IRExpr._translate(c_stmt.Ist.AbiHint.nia, irsb)

    def __str__(self):
        return "====== AbiHint(0x%s, %d, %s) ======" % (self.base, self.len, self.nia)


class Put(IRStmt):
    """
    Write to a guest register, at a fixed offset in the guest state.
    """

    __slots__ = ['data', 'offset']

    def __init__(self, c_stmt, irsb):
        IRStmt.__init__(self, c_stmt, irsb)
        self.data = IRExpr._translate(c_stmt.Ist.Put.data, irsb)
        self.offset = c_stmt.Ist.Put.offset

    def __str__(self):
        return "PUT(%s) = %s" % (self.arch.translate_register_name(self.offset, self.data.result_size/8), self.data)


class PutI(IRStmt):
    """
    Write to a guest register, at a non-fixed offset in the guest state.
    """

    __slots__ = ['descr', 'ix', 'data', 'bias']

    def __init__(self, c_stmt, irsb):
        IRStmt.__init__(self, c_stmt, irsb)
        self.descr = IRRegArray(c_stmt.Ist.PutI.details.descr)
        self.ix = IRExpr._translate(c_stmt.Ist.PutI.details.ix, irsb)
        self.data = IRExpr._translate(c_stmt.Ist.PutI.details.data, irsb)
        self.bias = c_stmt.Ist.PutI.details.bias

    def __str__(self):
        return "PutI(%s)[%s,%d] = %s" % (self.descr, self.ix, self.bias, self.data)


class WrTmp(IRStmt):
    """
    Assign a value to a temporary.  Note that SSA rules require each tmp is only assigned to once.  IR sanity checking
    will reject any block containing a temporary which is not assigned to exactly once.
    """

    __slots__ = ['data', 'tmp']

    def __init__(self, c_stmt, irsb):
        IRStmt.__init__(self, c_stmt, irsb)

        self.data = IRExpr._translate(c_stmt.Ist.WrTmp.data, irsb)
        self.tmp = c_stmt.Ist.WrTmp.tmp

    def __str__(self):
        return "t%d = %s" % (self.tmp, self.data)


class Store(IRStmt):
    """
    Write a value to memory..
    """

    __slots__ = ['addr', 'data', 'end']

    def __init__(self, c_stmt, irsb):
        IRStmt.__init__(self, c_stmt, irsb)

        self.addr = IRExpr._translate(c_stmt.Ist.Store.addr, irsb)
        self.data = IRExpr._translate(c_stmt.Ist.Store.data, irsb)
        self.end = ints_to_enums[c_stmt.Ist.Store.end]

    @property
    def endness(self):
        return self.end

    def __str__(self):
        return "ST%s(%s) = %s" % (self.endness[-2:].lower(), self.addr, self.data)


class CAS(IRStmt):
    """
    an atomic compare-and-swap operation.
    """

    __slots__ = ['addr', 'dataLo', 'dataHi', 'expdLo', 'expdHi', 'oldLo', 'oldHi', 'end']

    def __init__(self, c_stmt, irsb):
        IRStmt.__init__(self, c_stmt, irsb)

        self.addr = IRExpr._translate(c_stmt.Ist.CAS.details.addr, irsb)
        self.dataLo = IRExpr._translate(c_stmt.Ist.CAS.details.dataLo, irsb)
        self.dataHi = IRExpr._translate(c_stmt.Ist.CAS.details.dataHi, irsb)
        self.expdLo = IRExpr._translate(c_stmt.Ist.CAS.details.expdLo, irsb)
        self.expdHi = IRExpr._translate(c_stmt.Ist.CAS.details.expdHi, irsb)
        self.oldLo = c_stmt.Ist.CAS.details.oldLo
        self.oldHi = c_stmt.Ist.CAS.details.oldHi
        self.end = ints_to_enums[c_stmt.Ist.CAS.details.end]

    @property
    def endness(self):
        return self.end

    def __str__(self):
        return "t(%s,%s) = CAS%s(%s :: (%s,%s)->(%s,%s))" % (
        self.oldLo, self.oldHi, self.end[-2:].lower(), self.addr, self.expdLo, self.expdHi, self.dataLo, self.dataHi)


class LLSC(IRStmt):
    """
     Either Load-Linked or Store-Conditional, depending on STOREDATA. If STOREDATA is NULL then this is a Load-Linked,
     else it is a Store-Conditional.
    """

    __slots__ = ['addr', 'storedata', 'result', 'end']

    def __init__(self, c_stmt, irsb):
        IRStmt.__init__(self, c_stmt, irsb)

        self.addr = IRExpr._translate(c_stmt.Ist.LLSC.addr, irsb)
        self.storedata = IRExpr._translate(c_stmt.Ist.LLSC.storedata, irsb)
        self.result = c_stmt.Ist.LLSC.result
        self.end = ints_to_enums[c_stmt.Ist.LLSC.end]

    @property
    def endness(self):
        return self.end

    def __str__(self):
        if self.storedata is None:
            return "t%d = LD%s-Linked(%s)" % (self.result, self.end[-2:].lower(), self.addr)
        else:
            return "t%d = ( ST%s-Cond(%s) = %s )" % (self.result, self.end[-2:].lower(), self.addr, self.storedata)


class MBE(IRStmt):

    __slots__ = ['event']

    def __init__(self, c_stmt, irsb):
        IRStmt.__init__(self, c_stmt, irsb)
        self.event = ints_to_enums[c_stmt.Ist.MBE.event]

    def __str__(self):
        return "MBusEvent-" + self.event


class Dirty(IRStmt):

    __slots__ = ['cee', 'guard', 'tmp', 'mFx', 'mAddr', 'mSize', 'nFxState', 'args']

    def __init__(self, c_stmt, irsb):
        IRStmt.__init__(self, c_stmt, irsb)
        self.cee = IRCallee(c_stmt.Ist.Dirty.details.cee)
        self.guard = IRExpr._translate(c_stmt.Ist.Dirty.details.guard, irsb)
        self.tmp = c_stmt.Ist.Dirty.details.tmp
        self.mFx = ints_to_enums[c_stmt.Ist.Dirty.details.mFx]
        self.mAddr = IRExpr._translate(c_stmt.Ist.Dirty.details.mAddr, irsb)
        self.mSize = c_stmt.Ist.Dirty.details.mSize
        self.nFxState = c_stmt.Ist.Dirty.details.nFxState

        args = []
        for i in range(20):
            a = c_stmt.Ist.Dirty.details.args[i]
            if a == ffi.NULL:
                break

            args.append(IRExpr._translate(a, irsb))
        self.args = tuple(args)

    def __str__(self):
        return "t%s = DIRTY %s %s ::: %s(%s)" % (
        self.tmp, self.guard, "TODO(effects)", self.cee, ','.join(str(a) for a in self.args))

    @property
    def child_expressions(self):
        expressions = sum((a.child_expressions for a in self.args), [])
        expressions.extend(self.args)
        expressions.append(self.guard)
        expressions.extend(self.guard.child_expressions)
        return expressions


class Exit(IRStmt):
    """
    A conditional exit from the middle of an IRSB.
    """

    __slots__ = ['guard', 'dst', 'offsIP', 'jk']

    def __init__(self, c_stmt, irsb):
        IRStmt.__init__(self, c_stmt, irsb)
        self.guard = IRExpr._translate(c_stmt.Ist.Exit.guard, irsb)
        self.dst = IRConst._translate(c_stmt.Ist.Exit.dst)
        self.offsIP = c_stmt.Ist.Exit.offsIP
        self.jk = ints_to_enums[c_stmt.Ist.Exit.jk]

    @property
    def jumpkind(self):
        return self.jk

    def __str__(self):
        return "if (%s) { PUT(%s) = %s; %s }" % (
        self.guard, self.arch.translate_register_name(self.offsIP), hex(self.dst.value), self.jumpkind)

    @property
    def child_expressions(self):
        return [self.guard, self.dst] + self.guard.child_expressions


class LoadG(IRStmt):
    """
    A guarded load.
    """

    __slots__ = ['addr', 'alt', 'guard', 'dst', 'cvt', 'end', 'cvt_types']

    def __init__(self, c_stmt, irsb):
        IRStmt.__init__(self, c_stmt, irsb)

        self.addr = IRExpr._translate(c_stmt.Ist.LoadG.details.addr, irsb)
        self.alt = IRExpr._translate(c_stmt.Ist.LoadG.details.alt, irsb)
        self.guard = IRExpr._translate(c_stmt.Ist.LoadG.details.guard, irsb)
        self.dst = c_stmt.Ist.LoadG.details.dst
        self.cvt = ints_to_enums[c_stmt.Ist.LoadG.details.cvt]

        self.end = ints_to_enums[c_stmt.Ist.LoadG.details.end]

        type_in = ffi.new('IRType *')
        type_out = ffi.new('IRType *')
        pvc.typeOfIRLoadGOp(c_stmt.Ist.LoadG.details.cvt, type_out, type_in)
        type_in = ffi.cast('int *', type_in)[0]
        type_out = ffi.cast('int *', type_out)[0]
        self.cvt_types = (ints_to_enums[type_in], ints_to_enums[type_out])

    @property
    def endness(self):
        return self.end

    def __str__(self):
        return "t%d = if (%s) %s(LD%s(%s)) else %s" % (
        self.dst, self.guard, self.cvt, self.end[-2:].lower(), self.addr, self.alt)


class StoreG(IRStmt):
    """
    A guarded store.
    """

    __slots__ = ['addr', 'data', 'guard', 'end']

    def __init__(self, c_stmt, irsb):
        IRStmt.__init__(self, c_stmt, irsb)

        self.addr = IRExpr._translate(c_stmt.Ist.StoreG.details.addr, irsb)
        self.data = IRExpr._translate(c_stmt.Ist.StoreG.details.data, irsb)
        self.guard = IRExpr._translate(c_stmt.Ist.StoreG.details.guard, irsb)
        self.end = ints_to_enums[c_stmt.Ist.StoreG.details.end]

    @property
    def endness(self):
        return self.end

    def __str__(self):
        return "if (%s) ST%s(%s) = %s" % (self.guard, self.end[-2:].lower(), self.addr, self.data)


from .expr import IRExpr
from .const import IRConst
from .enums import IRRegArray, ints_to_enums, enums_to_ints, IRCallee
from .errors import PyVEXError
from . import ffi, pvc

_tag_to_class = {
    enums_to_ints['Ist_NoOp']: NoOp,
    enums_to_ints['Ist_IMark']: IMark,
    enums_to_ints['Ist_AbiHint']: AbiHint,
    enums_to_ints['Ist_Put']: Put,
    enums_to_ints['Ist_PutI']: PutI,
    enums_to_ints['Ist_WrTmp']: WrTmp,
    enums_to_ints['Ist_Store']: Store,
    enums_to_ints['Ist_LoadG']: LoadG,
    enums_to_ints['Ist_StoreG']: StoreG,
    enums_to_ints['Ist_CAS']: CAS,
    enums_to_ints['Ist_LLSC']: LLSC,
    enums_to_ints['Ist_Dirty']: Dirty,
    enums_to_ints['Ist_MBE']: MBE,
    enums_to_ints['Ist_Exit']: Exit,
}



import collections
from . import pvc, ffi

_counts = collections.Counter()

class VEXObject(object):
    """
    The base class for Vex types.
    """

    __slots__ = [ ]

    # def __init__(self):
    #   print "CREATING:",type(self)
    #   _counts[type(self)] += 1

    # def __del__(self):
    #   print "DELETING:",type(self)
    #   _counts[type(self)] -= 1

class IRCallee(VEXObject):
    """
    Describes a helper function to call.
    """

    __slots__ = ['regparms', 'name', 'mcx_mask', 'addr']

    def __init__(self, callee):
        VEXObject.__init__(self)
        self.regparms = callee.regparms
        self.name = ffi.string(callee.name)
        self.mcx_mask = callee.mcx_mask
        self.addr = int(ffi.cast("unsigned long long", callee.mcx_mask))

    def __str__(self):
        return self.name


class IRRegArray(VEXObject):
    """
    A section of the guest state that we want te be able to index at run time, so as to be able to describe indexed or
    rotating register files on the guest.

    :ivar int base:     The offset into the state that this array starts
    :ivar str elemTy:   The types of the elements in this array, as VEX enum strings
    :ivar int nElems:   The number of elements in this array
    """

    __slots__ = ['base', 'elemTy', 'nElems']

    def __init__(self, arr):
        VEXObject.__init__(self)
        self.base = arr.base
        self.elemTy = ints_to_enums[arr.elemTy]
        self.nElems = arr.nElems

    def __str__(self):
        return "%s:%sx%d" % (self.base, self.elemTy[4:], self.nElems)

enums_to_ints = {_: getattr(pvc, _) for _ in dir(pvc) if hasattr(pvc, _) and isinstance(getattr(pvc, _), int)}
ints_to_enums = {getattr(pvc, _): _ for _ in dir(pvc) if hasattr(pvc, _) and isinstance(getattr(pvc, _), int)}
enum_IROp_fromstr = {_: enums_to_ints[_] for _ in enums_to_ints if _.startswith('Iop_')}
type_sizes = {
    'Ity_INVALID': None,
    'Ity_I1': 1,
    'Ity_I8': 8,
    'Ity_I16': 16,
    'Ity_I32': 32,
    'Ity_I64': 64,
    'Ity_I128': 128,
    'Ity_F16': 16,
    'Ity_F32': 32,
    'Ity_F64': 64,
    'Ity_F128': 128,
    'Ity_D32': 32,
    'Ity_D64': 64,
    'Ity_D128': 128,
    'Ity_V128': 128,
    'Ity_V256': 256
}

def _get_op_type(op):
    irsb = pvc.emptyIRSB()
    t = pvc.newIRTemp(irsb.tyenv, pvc.Ity_I8)
    e = pvc.IRExpr_Unop(enums_to_ints[op], pvc.IRExpr_RdTmp(t))
    return ints_to_enums[pvc.typeOfIRExpr(irsb.tyenv, e)]


_op_types = {_: _get_op_type(_) for _ in enums_to_ints if
             _.startswith('Iop_') and _ != 'Iop_INVALID' and _ != 'Iop_LAST'}

def typeOfIROp(op):
    return _op_types[op]


def vex_endness_from_string(endness_str):
    return getattr(pvc, endness_str)

def default_vex_archinfo():
    return {
        'hwcaps': 0,
        'endness': vex_endness_from_string('VexEndnessLE'),
        'ppc_icache_line_szB': 0,
        'ppc_dcbz_szB': 0,
        'ppc_dcbzl_szB': 0,
        'arm64_dMinLine_lg2_szB': 0,
        'arm64_iMinLine_lg2_szB': 0,
        'hwcache_info': {
            'num_levels': 0,
            'num_caches': 0,
            'caches': None,
            'icaches_maintain_coherence': True,
        },
        'x86_cr0': 0xffffffff
    }




import threading
from . import VEXObject

_libvex_lock = threading.Lock()

class IRSB(VEXObject):
    """
    The IRSB is the primary interface to pyvex. Constructing one of these will make a call into LibVEX to perform a
    translation.

    IRSB stands for *Intermediate Representation Super-Block*. An IRSB in VEX is a single-entry, multiple-exit code
    block.

    :ivar arch:             The architecture this block is lifted under
    :vartype arch:          :class:`archinfo.Arch`
    :ivar statements:       The statements in this block
    :vartype statements:    list of :class:`IRStmt`
    :ivar next:             The expression for the default exit target of this block
    :vartype next:          :class:`IRExpr`
    :ivar int offsIP:       The offset of the instruction pointer in the VEX guest state
    :ivar int stmts_used:   The number of statements in this IRSB
    :ivar str jumpkind:     The type of this block's default jump (call, boring, syscall, etc) as a VEX enum string
    :ivar bool direct_next: Whether this block ends with a direct (not indirect) jump or branch
    """

    __slots__ = ['c_irsb', 'arch', 'statements', 'next', 'tyenv', 'offsIP', 'stmts_used', 'jumpkind', '_addr',
                 'direct_next'
                 ]

    def __init__(self, data, mem_addr, arch, num_inst=None, num_bytes=None, bytes_offset=0,
                 traceflags=0):  # pylint:disable=redefined-builtin
        """
        :param data:            The bytes to lift. Can be either a string of bytes or a cffi buffer object.
        :type data:             str or bytes or cffi.FFI.CData
        :param int mem_addr:    The address to lift the data at.
        :param arch:            The architecture to lift the data as.
        :type arch:             :class:`archinfo.Arch`
        :param num_inst:        The maximum number of instructions to lift. Max 99. (See note below)
        :param num_bytes:       The maximum number of bytes to use. Max 400.
        :param bytes_offset:    The offset into `data` to start lifting at.
        :param traceflags:      The libVEX traceflags, controlling VEX debug prints.

        .. note:: Explicitly specifying the number of instructions to lift (`num_inst`) may not always work
                  exactly as expected. For example, on MIPS, it is meaningless to lift a branch or jump
                  instruction without its delay slot. VEX attempts to Do The Right Thing by possibly decoding
                  fewer instructions than requested. Specifically, this means that lifting a branch or jump
                  on MIPS as a single instruction (`num_inst=1`) will result in an empty IRSB, and subsequent
                  attempts to run this block will raise `SimIRSBError('Empty IRSB passed to SimIRSB.')`.
        """
        try:
            _libvex_lock.acquire()

            VEXObject.__init__(self)

            if isinstance(data, (str, bytes)):
                num_bytes = len(data) if num_bytes is None else num_bytes
                c_bytes = ffi.new('char [%d]' % (num_bytes + 8), data + '\0' * 8)
            else:
                if not num_bytes:
                    raise PyVEXError("C-backed bytes must have the length specified by num_bytes")
                c_bytes = data

            if num_bytes == 0:
                raise PyVEXError("No bytes provided")
            pvc.vta.traceflags = traceflags

            vex_arch = getattr(pvc, arch.vex_arch)

            arch.vex_archinfo['hwcache_info']['caches'] = ffi.NULL

            if num_inst is not None:
                c_irsb = pvc.vex_block_inst(vex_arch, arch.vex_archinfo, c_bytes + bytes_offset, mem_addr, num_inst)
            else:
                c_irsb = pvc.vex_block_bytes(vex_arch, arch.vex_archinfo, c_bytes + bytes_offset, mem_addr, num_bytes,
                                             1)

            if c_irsb == ffi.NULL:
                raise PyVEXError(ffi.string(pvc.last_error) if pvc.last_error != ffi.NULL else "unknown error")

            # We must use a pickle value, CData objects are not pickeable so not ffi.NULL
            arch.vex_archinfo['hwcache_info']['caches'] = None

            self.c_irsb = c_irsb
            self.arch = arch
            self.statements = [stmt.IRStmt._translate(c_irsb.stmts[i], self) for i in xrange(c_irsb.stmts_used)]
            self.next = expr.IRExpr._translate(c_irsb.next, self)
            self.tyenv = IRTypeEnv(c_irsb.tyenv)
            self.offsIP = c_irsb.offsIP
            self.stmts_used = c_irsb.stmts_used
            self.jumpkind = ints_to_enums[c_irsb.jumpkind]

            self._addr = mem_addr
            self.direct_next = self._is_defaultexit_direct_jump()

            del self.c_irsb
        finally:
            _libvex_lock.release()

    def pp(self):
        """
        Pretty-print the IRSB to stdout.
        """
        print self._pp_str()

    def _pp_str(self):
        """
        Return the pretty-printed IRSB.

        :rtype: str
        """
        sa = []
        sa.append("IRSB {")
        sa.append("   %s" % self.tyenv)
        sa.append("")
        for i, s in enumerate(self.statements):
            sa.append("   %02d | %s" % (i, s))
        sa.append(
            "   NEXT: PUT(%s) = %s; %s" % (self.arch.translate_register_name(self.offsIP), self.next, self.jumpkind))
        sa.append("}")
        return '\n'.join(sa)

    @property
    def expressions(self):
        """
        A list of all expressions contained in the IRSB.
        """
        expressions = []
        for s in self.statements:
            expressions.extend(s.expressions)
        expressions.append(self.next)
        return expressions

    @property
    def instructions(self):
        """
        The number of instructions in this block
        """
        return len([s.addr for s in self.statements if isinstance(s, stmt.IMark)])

    @property
    def size(self):
        """
        The size of this block, in bytes
        """
        return sum([s.len for s in self.statements if isinstance(s, stmt.IMark)])

    @property
    def operations(self):
        """
        A list of all operations done by the IRSB, as libVEX enum names
        """
        ops = []
        for e in self.expressions:
            if hasattr(e, 'op'):
                ops.append(e.op)
        return ops

    @property
    def all_constants(self):
        """
        Returns all constants in the block (including incrementing of the program counter) as :class:`pyvex.const.IRConst`.
        """
        return sum((e.constants for e in self.expressions), [])

    @property
    def constants(self):
        """
        The constants (excluding updates of the program counter) in the IRSB as :class:`pyvex.const.IRConst`.
        """
        return sum(
            (s.constants for s in self.statements if not (isinstance(s, stmt.Put) and s.offset == self.offsIP)), [])

    @property
    def constant_jump_targets(self):
        """
        A set of the static jump targets of the basic block.
        """
        exits = set()
        for s in self.statements:
            if isinstance(s, stmt.Exit):
                exits.add(s.dst.value)

        default_target = self._get_defaultexit_target()
        if default_target is not None:
            exits.add(default_target)

        return exits

    @property
    def constant_jump_targets_and_jumpkinds(self):
        """
        A dict of the static jump targets of the basic block to their jumpkind.
        """
        exits = dict()
        for s in self.statements:
            if isinstance(s, stmt.Exit):
                exits[s.dst.value] = s.jumpkind

        default_target = self._get_defaultexit_target()
        if default_target is not None:
            exits[default_target] = self.jumpkind

        return exits


    def _get_defaultexit_target(self):
        """
        The default exit target, if it is constant, or None.
        """
        if isinstance(self.next, expr.Const):
            return self.next.con.value

        if not isinstance(self.next, expr.RdTmp):
            raise PyVEXError("unexpected self.next type: %s", self.next.__class__.__name__)

        tmp_next = self.next.tmp
        reg_next = None
        reg_next_size = None
        for stat in reversed(self.statements):
            if isinstance(stat, stmt.WrTmp) and stat.tmp == tmp_next:
                data = stat.data
            elif isinstance(stat, stmt.Put) and stat.offset == reg_next:
                data = stat.data
                if data.result_size != reg_next_size:
                    return None
            elif isinstance(stat, stmt.LoadG) and stat.dst == tmp_next:
                return None
            else:
                continue

            if isinstance(data, expr.Const):
                return data.con.value
            elif isinstance(data, expr.RdTmp):
                tmp_next = data.tmp
                reg_next = None
            elif isinstance(data, expr.Get):
                tmp_next = None
                reg_next = data.offset
                reg_next_size = data.result_size
            else:
                return None

        if tmp_next is not None:
            raise PyVEXError('Malformed IRSB at address #%x. Please report to Fish.' % self._addr)
        return None

    def _is_defaultexit_direct_jump(self):
        """
        Checks if the default of this IRSB a direct jump or not.
        """
        if not (self.jumpkind == 'Ijk_Boring' or self.jumpkind == 'Ijk_Call'):
            return False

        target = self._get_defaultexit_target()
        return target is not None


class IRTypeEnv(VEXObject):
    """
    An IR type environment.

    :ivar types:        A list of the types of all the temporaries in this block as VEX enum strings.
                        `types[3]` is the type of t3.
    :vartype types:     list of str
    """

    __slots__ = [ 'types', 'types_used' ]

    def __init__(self, tyenv):
        VEXObject.__init__(self)
        self.types = [ints_to_enums[tyenv.types[t]] for t in xrange(tyenv.types_used)]
        self.types_used = tyenv.types_used

    def __str__(self):
        return ' '.join(("t%d:%s" % (i, t)) for i, t in enumerate(self.types))

from . import expr, stmt, ffi, pvc
from .enums import ints_to_enums
from .errors import PyVEXError
